;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC GUEZE Drawing Editor
; Distribution events management
;
;
; (LAMBIC-USER::START-SESSION LAMBIC-USER::*GUEZE* "FIRST SESSION")
; 
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic-user)

(defmethod gueze-growl (title message)
  (growl "Gueze notification" title message "Lambic Gueze"))

;;;;;;;;;;;;;;;;;;;;;;;
; DISTRIBUTION EVENTS ;
;;;;;;;;;;;;;;;;;;;;;;;

; Generate next session index
(defun generate-next-session-index ()
  (setf *editor-index* (1+ *editor-index*)))

; Start session
(defmethod start-session (this-editor session-name)
  (setf (in-session this-editor) session-name)
  (setf *editor-index* 0)
  (setf (editor-index this-editor) (generate-next-session-index))
  (setf (is-leader this-editor) t))

; Display session error
(defun display-session-error ()
  (capi:popup-confirmer nil 
                        "You are not leading a session at the moment"
                        :callback-type :none
                        :ok-button "Close"
                        :no-button nil
                        :cancel-button nil
                        :value-function #'(c2cl:lambda (dummy)
                                            (declare (ignore dummy))
                                            nil)))

(defmethod display-joining-error (peer-editor-name)
  (capi:popup-confirmer nil 
                        (format nil "~d has refused to join session" peer-editor-name)
                        :callback-type :none
                        :ok-button "Close"
                        :no-button nil
                        :cancel-button nil
                        :value-function #'(c2cl:lambda (dummy)
                                            (declare (ignore dummy))
                                            nil)))

; Add editor (with username given as parameter) to session
(defmethod add-editor ((this-editor gueze-editor) username)
  (if (in-session this-editor)
      (let ((peer-editor (find-if (lambda (editor)
                                    (equal (in-actor-of editor (editor-username editor))
                                           username))
                                  (group-peers this-editor))))
        (when peer-editor
          (join-session this-editor peer-editor nil)))
    (display-session-error)))


(defmethod join-session ((this-editor gueze-editor) peer-editor reconnectedp)
  (let* ((new-index (generate-next-session-index))
         (session (in-session this-editor))
         (joinedp (accept-join-request peer-editor session new-index)))
    (if joinedp
        (peer-joined this-editor peer-editor reconnectedp)
      (display-joining-error (editor-username peer-editor)))))



#|
         (other-peers-records (loop for figure in (editor-figures this-editor)
                                    when (not (eql (figure-owner figure) peer-editor))
                                    collect (generate-figure-record figure)))
         (session (in-session this-editor))
         (peer-records (accept-join-request peer-editor session new-index 
                                            other-peers-records)))
    (if peer-records
        (progn
          (peer-joined this-editor peer-editor reconnectedp)
          (update-figures this-editor peer-records))
      (display-joining-error (editor-username peer-editor)))))
|#

; Join a session with index given by session's leader.
; In case of accepting request, return editor's figures.
(defmethod accept-join-request ((this-editor gueze-editor) session new-index)
  (debug-print "accept-join-request" this-editor)
  (let ((joinp (or (equal session (in-session this-editor))
                   (capi:popup-confirmer nil 
                                         "Join collaborative session?"
                                         :callback-type :none
                                         :ok-button "Confirm"
                                         :no-button "Cancel"
                                         :cancel-button nil
                                         :value-function #'(c2cl:lambda (dummy)
                                                             (declare (ignore dummy))
                                                             t)))))
    (when joinp
      (setf (editor-index this-editor) new-index)
      (setf (in-session this-editor) session))))

#|
      (update-figures this-editor other-peers-records)
      (generate-figure-records this-editor))))
|#


; Get a record with the info of each figure.
(defun generate-figure-records (editor)
  (loop for figure in (editor-figures editor)
        when (eql editor (figure-owner figure))
        collect (generate-figure-record figure)))

; Generate a GUEZE-FIGURE-RECORD from a GUEZE-FIGURE
(defmethod generate-figure-record ((figure gueze-figure))
  (capi:with-geometry figure
    (make-instance 'gueze-figure-record
                   :class (class-of figure) :name (figure-name figure)
                   :version (figure-version figure) :owner (figure-owner figure) 
                   :x capi:%x% :y capi:%y%
                   :w capi:%width% :h capi:%height%
                   :color (getf (capi:pinboard-object-graphics-args 
                                 figure) :foreground))))

; Generate a GUEZE-DRAWN-FIGURE-RECORD from a GUEZE-DRAWN-FIGURE
(defmethod generate-figure-record ((figure gueze-drawn-figure))
  (make-instance 'gueze-drawn-figure-record
                 :name (figure-name figure) :owner (figure-owner figure)
                 :version (figure-version figure) :line (line figure)
                 :color (getf (capi:pinboard-object-graphics-args 
                               figure) :foreground)))

;;;;;;;;;;;;;;;;;;;;;;;;;
; CONNECTION MANAGEMENT ;
;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;
; PEER JOINED ;
;;;;;;;;;;;;;;;

(defgroupgeneric peer-joined (this-editor peer-editor rejoinedp)
  (:predicates 'eql 'new-peer))

(defgroupmethod peer-joined ((this-editor gueze-editor) peer-editor rejoinedp)
  (declare (ignore peer-editor rejoined))
  (:propagate ((each this-editor)))
  (call-next-method))

(defmethod peer-joined ((this-editor gueze-editor) peer-editor rejoinedp)
  (:when (eql this-editor peer-editor) (new-peer rejoinedp))
  (gueze-growl "Lambic Gueze" "You have joined the session"))
;  (debug-print (generate-figure-records this-editor)
;  (update-figures this-editor (generate-figure-records this-editor)))

(defmethod peer-joined ((this-editor gueze-editor) peer-editor rejoinedp)
  (:when (new-peer rejoinedp))
  (gueze-growl "Lambic Gueze" 
               (format nil "~d has joined the session" 
                       (editor-username peer-editor))))
;  (debug-print "(generate-figure-records this-editor)" (generate-figure-records this-editor))
;  (update-figures this-editor (generate-figure-records this-editor)))

#|
(defmethod peer-joined ((this-editor gueze-editor) peer-editor rejoinedp)
  ;(cache-functions peer-editor)
  (debug-print "ready" "to update"))
;  (update-figures this-editor (generate-figure-records this-editor)))
|#

(defmethod new-peer (rejoinedp)
  (not rejoinedp))

;;;;;;;;;;;;;;;;;;;;;
; PEER DISCONNECTED ;
;;;;;;;;;;;;;;;;;;;;;

(defmethod disconnected-leader (this-editor peer-editor)
  (and (in-session this-editor)
       (is-leader peer-editor)))

(defmethod local-leader (this-editor)
  (and (in-session this-editor)
       (is-leader this-editor)))

(defgeneric peer-disconnected (this-editor peer-editor)
  (:predicates 'local-leader 'disconnected-leader))

(defmethod peer-disconnected ((this-editor gueze-editor) peer-editor)
  (:when (local-leader this-editor))
  (deselect-figures this-editor peer-editor)
  (delete-figures this-editor peer-editor)
  (call-next-method))

(defmethod peer-disconnected ((this-editor gueze-editor) peer-editor)
  (:when (disconnected-leader this-editor peer-editor))
  (if (became-leader this-editor peer-editor)
      (setf (is-leader this-editor) t)
    (update-figures this-editor (generate-figure-records this-editor)))
  (call-next-method))

(defmethod peer-disconnected ((this-editor gueze-editor) peer-editor)
  (gueze-growl "Lambic Gueze" 
               (format nil "~d is offline" (editor-username peer-editor))))

(defmethod deselect-peer-figures (this-editor peer-editor)
  (loop for figure in (editor-figures this-editor)
        if (eql (figure-user figure) peer-editor)
        do (deselect-figure this-editor figure)))

(defmethod became-leader (this-editor)
  (loop for peer-editor in (group-peers this-editor)
        with index = (editor-index this-editor)
        if (< (editor-index peer-editor) index)
        return nil
        finally return t))

;;;;;;;;;;;;;;;;;;;;
; PEER RECONNECTED ;
;;;;;;;;;;;;;;;;;;;;

(defgeneric peer-reconnected (this-editor peer-editor)
  (:predicates 'local-leader))

(defmethod peer-reconnected ((this-editor gueze-editor) peer-editor)
  (:when (local-leader this-editor))
  (join-session this-editor peer-editor t)
  (call-next-method))

(defmethod peer-reconnected ((this-editor gueze-editor) peer-editor)
  (gueze-growl "Lambic Gueze" 
               (format nil "~d is online" (editor-username peer-editor))))

;;;;;;;;;;;;;;;;;;;;;;;
; LOCAL DISCONNECTION ;
;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric local-disconnection (this-editor)
  (:predicates 'is-leader))

(defmethod local-disconnection ((this-editor gueze-editor))
  (:when (is-leader this-editor))
  (lose-leadership this-editor)
  (call-next-method))

(defmethod local-disconnection ((this-editor gueze-editor))
  (gueze-growl "Lambic Gueze" "You are offline"))

(defmethod lose-leadership (this-editor)
  (setf (editor-index this-editor) -1)
  (setf (is-leader this-editor) nil))

;;;;;;;;;;;;;;;;;;;;;;
; LOCAL RECONNECTION ;
;;;;;;;;;;;;;;;;;;;;;;

(defgeneric local-reconnection (this-editor)
  (:predicates 'is-leader))

(defmethod local-reconnection ((this-editor gueze-editor))
  (gueze-growl "Lambic Gueze" "You are online"))


  