;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC GUEZE Drawing Editor
; Group behaviour
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic-user)

; Figure as record
(defprotoclass gueze-figure-record ()
  ((class :initarg :class :accessor figure-record-class)
   (name :initarg :name :initform 'gueze-figure :accessor figure-record-name)
   (version :initarg :version :initform 0 :accessor figure-record-version)
   (owner :initarg :owner :accessor figure-record-owner)
   (user :initarg :user :initform nil :accessor figure-record-user)
   (x :initarg :x :initform 0 :accessor figure-record-x)
   (y :initarg :y :initform 0 :accessor figure-record-y)
   (w :initarg :w :initform 0 :accessor figure-record-w)
   (h :initarg :h :initform 0 :accessor figure-record-h)
   (color :initarg :color :initform :black :accessor figure-record-color)))

; Drawn figure as record
(defprotoclass gueze-drawn-figure-record (gueze-figure-record)
  ((line :initarg :line :accessor figure-record-line)))

;;;;;;;;;;;;;;
; ADD FIGURE ;
;;;;;;;;;;;;;;

(defmethod local-figure (figures)
  (find-if (lambda (figure)
             (co-located figure (figure-owner figure)))
           figures))

(defmethod not-leader-owner (editor (figure gueze-figure))
  (and (in-session editor)
       (not (is-leader editor))
       (eql editor (figure-owner figure))))

(defmethod not-leader-owner (editor (figure-records list))
  (when figure-records
    (not-leader-owner editor (car figure-records))))

(defmethod not-leader-owner (editor figure-record)
  (and (in-session editor)
       (not (is-leader editor))
       (eql editor (figure-record-owner figure-record))))

(defmethod session-leader (editor)
  (and (in-session editor)
       (is-leader editor)))

(defgroupgeneric add-figure (editor figure-record)
  (:group-predicates 'session-leader 'not-leader-owner))

(defgroupmethod add-figure ((editor gueze-editor) figure-record)
  (:when (session-leader editor))
  (:propagate ((each editor))
   :reduce-results-to 'local-figure
   :negligible t)
  (call-next-method))

(defgroupmethod add-figure ((editor gueze-editor) figure-record)
  (:when (not-leader-owner editor figure-record))
  (:propagate ((session-leader editor))
   :reduce-results-to 'local-figure
   :negligible t)
  (call-next-method))

; Add a figure to the editor
(defmethod add-figure ((editor gueze-editor) figure-record)
  (with-slots (class name owner version x y w h color)
      figure-record
;    (format *screen-output* "class: ~d~% name: ~d~% owner: ~d~% version: ~d~% x: ~d~% y: ~d~% w: ~d~% h: ~d~% color: ~d~%" class name owner version x y w h color)
    (let ((figure (make-instance class :owner owner :name name :version version
                                 :x x :y y :width w :height h
                                 :graphics-args (list :background :transparent
                                                      :foreground color))))
      
      (capi:apply-in-pane-process editor 
                                  #'(c2cl:lambda (editor figure)
                                      (capi:manipulate-pinboard editor figure :add-top))
                                  editor figure)
      (push figure (editor-figures editor))
      figure)))

; Add a drawn figure to editor
(defmethod add-figure ((editor gueze-editor) (figure-record gueze-drawn-figure-record))
  (with-slots (owner name version line color)
      figure-record
    (multiple-value-bind (tl-x tl-y br-x br-y)
        (eval `(gp:get-enclosing-rectangle ,@line))
      (let* ((x (- tl-x 5)) 
             (y (- tl-y 5))
             (width (+ 10 (- br-x tl-x))) 
             (height (+ 10 (- br-y tl-y)))
             (figure (make-instance 'gueze-drawn-figure :owner owner :name name 
                                    :version version :line line :x x :y y
                                    :width width :height height
                                    :graphics-args (list :background :transparent
                                                         :foreground color))))
        (setf (status figure) (make-figure-status :x x :y y
                                                  :width width :height height))
        (push figure (editor-figures editor))
        (capi:apply-in-pane-process editor 
                                  #'(c2cl:lambda (editor figure)
                                      (capi:manipulate-pinboard editor figure :add-top))
                                  editor figure)
        figure))))

;;;;;;;;;;;;;;;;;;;;
; SELECTING FIGURE ;
;;;;;;;;;;;;;;;;;;;;

(defmethod co-located ((editor gueze-editor) (figure gueze-figure))
  (find figure (editor-figures editor)))

(defmethod co-located (editor figure)
  (equal (actor-of editor) (actor-of figure)))

(defmethod session-peer (editor)
  (not (session-leader editor)))

(defmethod same-name (figure)
  (equal (figure-name (get-original-binding figure))
         (figure-name figure)))

(defgroupgeneric select-figure (editor figure user-editor)
  (:group-predicates 'session-leader 'session-peer))

(defgroupmethod select-figure ((editor gueze-editor) (figure gueze-figure) user-editor)
  (:when (session-peer editor))
  (:propagate ((session-leader editor) (co-located editor figure) (same-name figure)) 
   :negligible t)
  (call-next-method))

(defgroupmethod select-figure ((editor gueze-editor) (figure gueze-figure) user-editor)
  (:when (session-leader editor))
  (:propagate ((each editor) (co-located editor figure) (same-name figure)) :negligible t)
  ;(debug-print "selecting figure: (figure-user figure)" (figure-user figure))
  (if (not (figure-user figure))
      (call-next-method)))

; Select figure
(defmethod select-figure ((editor gueze-editor) (figure gueze-figure) user-editor)
  (setf (figure-user figure) user-editor)
  (capi:apply-in-pane-process
   editor 
   #'(c2cl:lambda (figure)
       (gp:invalidate-rectangle (capi:pinboard-object-pinboard figure)))
   figure)
  figure)

(defun display-selection-error ()
  (capi:popup-confirmer nil 
                        "You cannot use this figure at the moment"
                        :callback-type :none
                        :ok-button "Close"
                        :no-button nil
                        :cancel-button nil
                        :value-function #'(c2cl:lambda (dummy)
                                            (declare (ignore dummy))
                                            nil)))

; Get figure's coordinates
(defun get-coords (figure)
  (capi:with-geometry figure
    (list capi:%x% capi:%y% (+ capi:%x% capi:%width%) (+ capi:%y% capi:%height%))))


; SELECTING ACTIONS
; Invoked on primary button motion
(defun draw-selection-rectangle (editor x y)
  (let ((selected-rectangle (selection-rectangle editor)))
    (if selected-rectangle
        (multiple-value-bind (ox oy)
            (capi:pinboard-pane-position selected-rectangle)
          (setf (capi:pinboard-pane-size selected-rectangle)
                (values (- x ox) (- y oy)))
          (dolist (figure (editor-figures editor))
            (let ((coords (get-coords selected-rectangle)))
              (capi:with-geometry figure
                (if (or (gp:inside-rectangle coords capi:%x% capi:%y%)
                        (gp:inside-rectangle coords capi:%x%
                                             (+ capi:%y% capi:%height%)))
                    (when (not (figure-user figure))
                      (select-figure editor figure editor))
                  (when (figure-user figure)
                    (deselect-figure editor figure)))))))
      (progn
        (setf selected-rectangle
              (make-instance 'capi:rectangle
                             :x x :y y :external-min-width 0 :external-min-height 0
                             :visible-min-width 0 :visible-min-height 0
                             :internal-min-width 0 :internal-min-width 0
                             :graphics-args '(:foreground :selection-color)
                             :filled nil))
        (setf (selection-rectangle editor) selected-rectangle)
        (capi:manipulate-pinboard editor selected-rectangle :add-top)))))

; Invoked on primary button released
(defun end-drawing-selection-rectangle (editor)
  (capi:manipulate-pinboard editor (selection-rectangle editor) :delete)
  (setf (selection-rectangle editor) nil))


;;;;;;;;;;;;;;;;;;;;;;
; DESELECTING FIGURE ;
;;;;;;;;;;;;;;;;;;;;;;

; Deselect editor's figures used by editor given as parameter.
(defmethod deselect-figures ((editor gueze-editor) user-editor)
  (dolist (figure (editor-figures editor))
    (if (eql (figure-user figure) user-editor)
        (deselect-figure editor figure))))

(defgroupgeneric deselect-figure (editor figure)
  (:group-predicates 'session-leader 'session-peer))

(defgroupmethod deselect-figure ((editor gueze-editor) (figure gueze-figure))
  (:when (session-peer editor))
  (:propagate ((session-leader editor) (co-located editor figure) (same-name figure))
   :negligible t)
  (call-next-method))

(defgroupmethod deselect-figure ((editor gueze-editor) (figure gueze-figure))
  (:when (session-leader editor))
  (:propagate ((each editor) (co-located editor figure) (same-name figure)) :negligible t)
  (call-next-method))

; Deselect figure
(defmethod deselect-figure ((editor gueze-editor) (figure gueze-figure))
  (setf (figure-user figure) nil)
  (capi:apply-in-pane-process
   editor 
   #'(c2cl:lambda (figure)
       (gp:invalidate-rectangle (capi:pinboard-object-pinboard figure)))
   figure))

;;;;;;;;;;;;;;;;;;;
; PAINTING FIGURE ;
;;;;;;;;;;;;;;;;;;;

(defmethod same-selection ((figure gueze-figure) (editor gueze-editor))
  (peer-equal (figure-user (get-original-binding figure))
              (figure-user figure)))

(defmethod same-selection (figure editor)
  nil)

#|
(defmethod same-selection (figure editor)
  (debug-print "figure in same-selection" figure)
  (debug-print "editor in same-selection" editor)
  (debug-print "(this editor)" (this editor))
  (debug-print "(co-located editor figure)" (co-located editor figure))
  (debug-print "(eql (figure-user (get-original-binding figure)) (figure-user figure))"
               (eql (figure-user (get-original-binding figure))
                    (figure-user figure)))
  (and (this editor)
       (co-located editor figure)
       (eql (figure-user (get-original-binding figure))
            (figure-user figure))))
|#

#|
(defgroupgeneric paint-figure (editor figure color)
  (:group-predicates 'session-leader 'session-peer))

(defgroupmethod paint-figure ((editor gueze-editor) (figure gueze-figure) color)
  (:when (session-peer editor))
  (:propagate ((session-leader editor) (co-located editor figure) (same-name figure)) :negligible t)
  (call-next-method))

(defgroupmethod paint-figure ((editor gueze-editor) (figure gueze-figure) color)
  (:when (session-leader editor))
  (:propagate ((each editor) (co-located editor figure) (same-name figure)) :negligible t)
  (call-next-method))
|#

(defgroupgeneric paint-figure (editor figure color)
  (:group-predicates 'in-session))

(defgroupmethod paint-figure ((editor gueze-editor) (figure gueze-figure) color)
  (:when (in-session editor))
  (:propagate ((each editor) (co-located editor figure) (same-name figure)) :negligible t)
  (call-next-method))

(defgroupmethod paint-figure ((editor gueze-editor) (figure gueze-figure) color)
  (:propagate ((same-selection figure editor)))
  (call-next-method))

; base peer behaviour
(defmethod paint-figure ((editor gueze-editor) (figure gueze-figure) color)
  (setf (figure-version figure) (1+ (figure-version figure)))
  (capi:apply-in-pane-process
   editor 
   #'(c2cl:lambda (figure color)
       (setf (getf (capi:pinboard-object-graphics-args figure) :foreground) color))
   figure color))

;;;;;;;;;;;;;;;;;;;
; RESIZING FIGURE ;
;;;;;;;;;;;;;;;;;;;

; Display resize cursor
(defun display-resize-cursor (editor)
  (unless (position *tl-finger-cursor* (capi:layout-description editor))
    (capi:with-geometry editor
      (let ((xo (+ capi:%x% capi:%width%)) 
            (yo (+ capi:%y% capi:%height%)))
        (dolist (figure (editor-figures editor))
          (when (eql (figure-user figure) editor)
            (multiple-value-bind (x y)
                (capi:pinboard-pane-position figure)
              (when (< x xo) (setf xo x))
              (when (< y yo) (setf yo y)))))
        (capi:manipulate-pinboard editor *tl-finger-cursor* :add-top)
        (setf (capi:pinboard-pane-position *tl-finger-cursor*)
              (values (- xo 23) (- yo 16)))))))


; Invoked on primary button pressed 
(defun start-resizing-figure (editor x y)
  (setf (drag-status editor) (make-drag-status :x x :y y))
  (display-resize-cursor editor))

(defgroupgeneric resize-figure (editor figure x y)
  (:group-predicates 'in-session))

#|
(defgroupgeneric resize-figure (editor figure x y)
  (:group-predicates 'session-leader 'session-peer))

(defgroupmethod resize-figure ((editor gueze-editor) (figure gueze-figure) x y)
  (:when (session-peer editor))
  (:propagate ((session-leader editor) (co-located editor figure) (same-name figure))
   :negligible t)
  (call-next-method))

(defgroupmethod resize-figure ((editor gueze-editor) (figure gueze-figure) x y)
  (:when (session-leader editor))
  (:propagate ((each editor) (co-located editor figure) (same-name figure)) :negligible t)
  (call-next-method))
|#

(defgroupmethod resize-figure ((editor gueze-editor) (figure gueze-figure) x y)
  (:when (in-session editor))
  (:propagate ((each editor) (co-located editor figure) (same-name figure)) :negligible t)
  (call-next-method))

(defgroupmethod resize-figure ((editor gueze-editor) (figure gueze-figure) x y)
  (:propagate ((same-selection figure editor)))
  (call-next-method))

; base peer behaviour
(defmethod resize-figure ((editor gueze-editor) (figure gueze-figure) x y)
  (setf (figure-version figure) (1+ (figure-version figure)))
  (capi:apply-in-pane-process
   editor 
   #'(c2cl:lambda (figure x y)
       (capi:with-geometry figure
         (setf (capi:pinboard-pane-size figure) (values (+ capi:%width% x)
                                                        (+ capi:%height% y)))))
   figure x y))

; Invoked on primary button released
(defun end-resizing-figure (editor)
  (capi:manipulate-pinboard editor *tl-finger-cursor* :delete))

;;;;;;;;;;;;;;;;;
; MOVING FIGURE ;
;;;;;;;;;;;;;;;;;

; Invoked on primary button pressed 
(defun start-moving-figure (editor x y)
  (setf (drag-status editor) (make-drag-status :x x :y y)))

#|
(defgroupgeneric move-figure (editor figure x y)
  (:group-predicates 'session-leader 'session-peer))

(defgroupmethod deselect-figure ((editor gueze-editor) (figure gueze-figure))
  (:when (session-peer editor))
  (:propagate ((session-leader editor) (co-located editor figure) (same-name figure))
   :negligible t)
  (call-next-method))

(defgroupmethod deselect-figure ((editor gueze-editor) (figure gueze-figure))
  (:when (session-leader editor))
  (:propagate ((each editor) (co-located editor figure) (same-name figure)) :negligible t)
  (call-next-method))
|#

(defgroupgeneric move-figure (editor figure x y)
  (:group-predicates 'in-session))

(defgroupmethod move-figure ((editor gueze-editor) (figure gueze-figure) x y)
  (:propagate ((same-selection figure editor)))
  (call-next-method))

(defgroupmethod move-figure ((editor gueze-editor) (figure gueze-figure) x y)
  (:when (in-session editor))
  (:propagate ((each editor) (co-located editor figure) (same-name figure)) :negligible t)
  (call-next-method))

(defmethod move-figure ((editor gueze-editor) (figure gueze-figure) x y)
  (setf (figure-version figure) (1+ (figure-version figure)))
  (capi:apply-in-pane-process
   editor 
   #'(c2cl:lambda (figure x y)
       (capi:with-geometry figure
         (setf (capi:pinboard-pane-position figure) (values (+ capi:%x% x)
                                                       (+ capi:%y% y)))))
   figure x y))

;;;;;;;;;;;;;;;;;;
; DRAWING FIGURE ;
;;;;;;;;;;;;;;;;;;

; Invoked on primary button pressed 
(defun start-drawing-line (editor x y)
  (setf (current-point editor) (cons x y)))

; Invoked on primary button motion
(defun draw-line (editor x y)
  (with-slots (current-line current-point) editor
    (when current-point
      (let ((old-x (car current-point))
            (old-y (cdr current-point)))
        (gp:draw-line editor old-x old-y x y :foreground (selected-color editor)
                      :line-end-style :round :line-joint-style :round :thickness 4)
        (setf current-line `(,old-x ,old-y ,x ,y ,@current-line))))
    (setf current-point (cons x y))))

; Invoked on primary button released
(defun end-drawing-line (editor)
  (let* ((current-line (current-line editor))
         (figure-record (make-instance 'gueze-drawn-figure-record
                                       :name (generate-figure-name 'gueze-drawn-figure-record)
                                       :owner editor :line current-line
                                       :color (selected-color editor)))
         (new-figure (add-figure editor figure-record)))
    ;(select-figure editor new-figure editor)
    (setf (current-line editor) nil
          (current-point editor) nil)))

;;;;;;;;;;;;;;;;;;;
; DELETING FIGURE ;
;;;;;;;;;;;;;;;;;;;

(defgroupgeneric delete-figure (editor figure)
  (:group-predicates 'session-leader 'not-leader-owner))

(defgroupmethod delete-figure ((editor gueze-editor) (figure gueze-figure))
  (:when (not-leader-owner editor figure))
  (:propagate ((session-leader editor)) :negligible t)
  (call-next-method))

(defgroupmethod delete-figure ((editor gueze-editor) (figure gueze-figure))
  (:when (session-leader editor))
  (:propagate ((each editor) (co-located editor figure)) :negligible t)
  (call-next-method))

; Delete figure from editor
(defmethod delete-figure ((editor gueze-editor) (figure gueze-figure))
  (setf (editor-figures editor) (remove figure (editor-figures editor)))
  (capi:apply-in-pane-process editor 
                              #'(c2cl:lambda (editor figure)
                                  (capi:manipulate-pinboard editor figure :delete))
                              editor figure))

(defmethod delete-figures (this-editor peer-editor)
  (loop for figure in (editor-figures this-editor)
        if (eql (figure-owner figure) peer-editor)
        do (delete-figure this-editor figure)))

;;;;;;;;;;;;;;;;;
; UPDATE FIGURE ;
;;;;;;;;;;;;;;;;;

(defun find-figure-with-record (editor figure-record)
  (loop for figure in (editor-figures editor)
        when (equal (figure-name figure) (figure-record-name figure-record))
        return figure))

#|
  (c2cl:find-if (lambda (figure)
             (equal (figure-name figure)
                    (figure-record-name figure-record)))
           (editor-figures editor)))
|#

(defmethod except-owner (figure-records editor)
  (not (eql editor (figure-record-owner (car figure-records)))))

(defgroupgeneric update-figures (editor figure-records)
  (:group-predicates 'session-leader 'not-leader-owner))

(defgroupmethod update-figures ((editor gueze-editor) figure-records)
  (:when (session-leader editor))
  (:propagate ((except-owner figure-records editor)))
  (call-next-method))

(defgroupmethod update-figures ((editor gueze-editor) figure-records)
  (:when (not-leader-owner editor figure-records))
  (:propagate ((session-leader editor)))
  (call-next-method))

(defmethod update-figures ((editor gueze-editor) figure-records)
  (loop for figure-record in figure-records
        do (update-figure editor figure-record)))

(defmethod update-figure ((editor gueze-editor) (figure-record gueze-drawn-figure-record))
  (let ((figure (find-figure-with-record editor figure-record)))
    (if figure
        (when (> (figure-record-version figure-record) (figure-version figure))
          (capi:apply-in-pane-process
           editor 
           #'(c2cl:lambda (figure figure-record)
               (with-slots (line version color)
                   figure-record
                 (setf (figure-version figure) version)
                 (setf (getf (capi:pinboard-object-graphics-args figure) :foreground) 
                       color)
                 (setf (line figure) line)))
           figure figure-record))
      (add-figure editor figure-record))))

(defmethod update-figure ((editor gueze-editor) (figure-record gueze-figure-record))
  (let ((figure (find-figure-with-record editor figure-record)))
    (debug-print "figure" figure)
    (if figure
        (when (> (figure-record-version figure-record) (figure-version figure))
          (capi:apply-in-pane-process
           editor 
           #'(c2cl:lambda (figure figure-record)
               (with-slots (x y w h version color)
                   figure-record
                 (setf (figure-version figure) version)
                 (setf (getf (capi:pinboard-object-graphics-args figure) :foreground)
                       color)
                 (set-pinboard-object-geometry figure :x x :y y :width w :height h)))
           figure figure-record))
      (add-figure editor figure-record))))