;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC GUEZE Drawing Editor
; GUI events management
;
; Execute:
;
#| 

(IN-PACKAGE :LAMBIC-USER)
(LAMBIC-GUEZE "Bob")

(IN-PACKAGE :LAMBIC-USER)
(LAMBIC-GUEZE "Alice")

(START-SESSION *GUEZE* "First session")
(ADD-EDITOR *GUEZE* "Bob")

|#
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic-user)

;;;;;;;;;;;;;;;
;  INTERFACE  ;
;;;;;;;;;;;;;;;

(defvar *pressed-figure* nil)

(capi:define-interface editor-interface ()
  ()
  (:panes
   (drag-layout
    capi:pinboard-layout
    :description (list (make-instance 'capi:rectangle
                                      :filled t :x 10 :y 5 :width 30 :height 30
                                      :graphics-args '(:foreground :black)
                                      :name 'gueze-rectangle)
                       (make-instance 'capi:ellipse
                                      :filled t :x 50 :y 5 :width 30 :height 30
                                      :graphics-args '(:foreground :black)
                                      :name 'gueze-ellipse))
    :background :transparent
    :input-model '(((:button-1 :press) drag-from))
    :visible-min-width 280)
   (color-pane
    capi:pinboard-layout
    :description (list (make-instance 'capi:rectangle
                                      :filled t :x 0 :y 5 :width 60 :height 30
                                      :graphics-args '(:foreground :black)))
    :input-model '(((:button-1 :press) option-pane-set-color))
    :background :transparent)
   (buttons-pane
    button-pinboard-pane
    :description (list (make-instance 'checkable-button
                                      :x 0 :y 5
                                      :checked-image-location #.(cl-user::current-pathname
                                                                 "images/selected-brush.png")
                                      :unchecked-image-location #.(cl-user::current-pathname
                                                                   "images/unselected-brush.png")
                                      :when-checked 'set-painting-mode)
                       (make-instance 'checkable-button
                                      :x 32 :y 5
                                      :checked-image-location #.(cl-user::current-pathname
                                                                 "images/selected-finger.png")
                                      :unchecked-image-location #.(cl-user::current-pathname
                                                                   "images/unselected-finger.png")
                                      :when-checked 'set-cursor-mode :checkedp t))
    :selected-button-position 0    
    :background :transparent :visible-min-width 280)
   (zoom-pane
    capi:option-pane
    :accessor color-chooser :items '(25 50 100 150 200 400 800)
    :selected-item 100 :print-function 'option-in-percentage 
    :selection-callback 'change-zoom :visible-max-width 70)
   (editor
    gueze-editor
    :drop-callback 'drop-shape-callback :accessor editor
    :draw-pinboard-objects :local-buffer :background :gray :cursor *br-finger-cursor*
    :horizontal-scroll t :vertical-scroll t :scroll-width 700 :scroll-height 500
    :fit-size-to-children nil
    :input-model `(((:button-1 :press) ,#'(c2cl:lambda (editor x y)
                                            (setf *pressed-figure* (get-figure-at-position editor x y))
                                            (primary-button-pressed editor *pressed-figure* x y)))
                   ((:button-1 :motion) ,#'(c2cl:lambda (editor x y)
                                             (primary-button-motion editor *pressed-figure* x y)))
                   ((:button-1 :release) ,#'(c2cl:lambda (editor x y)
                                              (declare (ignore x y))
                                              (primary-button-released editor *pressed-figure*)
                                              (setf *pressed-figure* nil)))
                   ((#\a :press) resize-key-pressed)
                   ((#\a :release) resize-key-released)
                   ((#\del :press) deletion-key-pressed))))
  (:layouts
   (bottom-layout
    capi:row-layout
    '(drag-layout color-pane buttons-pane zoom-pane)
    :adjust :center :uniform-size-p nil :visible-max-height 40
    :ratios '(nil nil nil nil) :y-adjust :center)
   (main-layout
    capi:column-layout
    '(editor bottom-layout)))
  (:default-initargs
   :layout 'main-layout :title "Lambic Gueze"
   :best-width 723 :best-height 582))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;
;   MAIN FUNCTION   ;
;;;;;;;;;;;;;;;;;;;;;

(defvar *gueze* nil)

(defun lambic-gueze (username)
  (let* ((interface (make-instance 'editor-interface))
         (editor (editor interface)))
    (setf *gueze* editor)
    (setf (editor-username editor) username)
    (capi:display interface)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;
;     CALLBACKS     ;
;;;;;;;;;;;;;;;;;;;;;

; Color button selected
(defun option-pane-set-color (pane x y)
  (let ((object (capi:pinboard-object-at-position pane x y)))
    (when object
      (let ((current-color (getf (capi:pinboard-object-graphics-args object) :foreground)))
        (multiple-value-bind (new-color successp)
            (capi:prompt-for-color "Colors" :color current-color)
          (when successp
            (setf (getf (capi:pinboard-object-graphics-args object) :foreground) new-color)
            (let ((editor (editor (capi:top-level-interface pane))))
              (setf (selected-color editor) new-color))))))))

; Finger button selected
(defun set-cursor-mode (interface)
  ; ENTER TO CURSOR MODE
  (setf (capi:simple-pane-cursor (editor interface)) *br-finger-cursor*)
  (setf (brush-activep (editor interface)) nil))

; Brush button selected
(defun set-painting-mode (interface)
  ; ENTER TO A PAINTING/DRAWING MODE
  (setf (capi:simple-pane-cursor (editor interface)) *brush-cursor*)
  (setf (brush-activep (editor interface)) t))

; Zoom option button changed
(defun change-zoom (zoom-percentage interface)
  (let* ((editor (editor interface))
         (all-figures (reverse (capi:layout-description editor)))
         (new-zoom-level (/ zoom-percentage 100))
         (delta (/ new-zoom-level (zoom-level editor))))
    (when (and all-figures (not (eql delta 1)))
      (dolist (figure all-figures)
        (capi:with-geometry figure
          (set-pinboard-object-geometry figure
                                        :x (* capi:%x% delta) :y (* capi:%y% delta)
                                        :width (* capi:%width% delta)
                                        :height (* capi:%height% delta))
          (when (eql figure (editor-canvas editor))
            (capi:set-horizontal-scroll-parameters editor :max-range capi:%width%)
            (capi:set-vertical-scroll-parameters editor :max-range capi:%height%))))
      (setf (zoom-level editor) new-zoom-level))))

; New shape selected
(defun drag-from (pane x y)
  (let ((object (capi:pinboard-object-at-position pane x y)))
    (when object
      (let ((value (capi:capi-object-name object)))
        (capi:drag-pane-object pane value)))))

; New shape droped
(defun drop-shape-callback (editor drop-object stage)
  (case stage
    (:formats
     (capi:set-drop-object-supported-formats drop-object '(:value)))
    (:enter
     (when (and (capi:drop-object-provides-format drop-object :value)
                (capi:drop-object-allows-drop-effect-p drop-object :copy))
       (setf (capi:drop-object-drop-effect drop-object) :copy)))
    (:drag
     (when (and (capi:drop-object-provides-format drop-object :value)
                (capi:drop-object-allows-drop-effect-p drop-object :copy))
       (setf (capi:drop-object-drop-effect drop-object) :copy)))
    (:drop
     (when (and (capi:drop-object-provides-format drop-object :value)
                (capi:drop-object-allows-drop-effect-p drop-object :copy))
       (capi:set-pane-focus editor)
       (deselect-figures editor editor)
       (let* ((value (capi:drop-object-get-object drop-object editor :value))
              (figure-record (make-instance 'gueze-figure-record
                                            :name (generate-figure-name value)
                                            :class value :owner editor :w 45 :h 45
                                            :x (capi:drop-object-pane-x drop-object)
                                            :y (capi:drop-object-pane-y drop-object)
                                            :color (selected-color editor)))
              (figure (add-figure editor figure-record)))
         (when figure
           (debug-print "ADDED FIGURE" figure)
           ;(select-figure editor figure editor)
           (setf (capi:drop-object-drop-effect drop-object) :copy)))))))

; Deletion key pressed
; Key: Delete
; TODO to remove from *distributable-objects* and bonjour
(defun deletion-key-pressed (editor x y key)
  (declare (ignore x y key type))
  (dolist (figure (editor-figures editor))
    (if (and (eql (figure-user figure) editor)
                  (eql (figure-owner figure) editor))
        (delete-figure editor figure)))
  (gp:invalidate-rectangle editor))

; Resizing key pressed
; Key: A
(defun resize-key-pressed (editor x y key)
  ; ENTER TO RESIZE MODE
  (declare (ignore x y key type))
  (setf (resize-activep editor) t)
  (when (drag-status editor)
    (display-resize-cursor editor)))

; Resizing key released
; Key: A
(defun resize-key-released (editor x y key) 
  ; LEAVE RESIZE MODE
  (declare (ignore x y key type))
  (setf (resize-activep editor) nil)
  (capi:manipulate-pinboard editor *tl-finger-cursor* :delete))

; Zoom option in percentage
(defun option-in-percentage (option)
  (format nil "~a%" option))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;
; GUI EVENTS ;
;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;
; PRIMARY-BUTTON-PRESSED ;
;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod deselecting (figure editor)
  (not figure))

(defmethod painting (figure editor)
  (and figure (brush-activep editor)
       (eql editor (figure-user figure))))

(defmethod selecting (figure editor)
  (and figure (not (figure-user figure))))

(defmethod figure-in-use (figure editor)
  (and figure
       (figure-user figure)
       (not (eql (figure-user figure) editor))))

(defgeneric primary-button-pressed (editor figure x y)
  (:predicates 'painting
               (moving (figure editor)
                 (and figure (not (brush-activep editor))
                      (not (resize-activep editor))))
               (resizing (figure editor)
                 (and figure (not (brush-activep editor))
                      (resize-activep editor)))
               (drawing (figure editor)
                 (and (not figure) (brush-activep editor)))
               'selecting
               'figure-in-use
               'deselecting))

(defmethod primary-button-pressed (editor figure x y)
  (declare (ignore x y))
  (:when (painting figure editor))
  (paint-figure editor figure (selected-color editor)))

(defmethod primary-button-pressed (editor figure x y)
  (:when (moving figure editor))
  (start-moving-figure editor x y))

(defmethod primary-button-pressed (editor figure x y)
  (:when (resizing figure editor))
  (start-resizing-figure editor x y))

(defmethod primary-button-pressed (editor figure x y)
  (:when (drawing figure editor))
  (start-drawing-line editor x y))

(defmethod primary-button-pressed (editor figure x y)
  (declare (ignore figure x y))
  (:when (selecting figure editor))
  (deselect-figures editor editor)
  (if (select-figure editor figure editor)
      (call-next-method)
    (display-selection-error)))

(defmethod primary-button-pressed (editor figure x y)
  (declare (ignore figure x y))
  (:when (figure-in-use figure editor))
  (display-selection-error))

(defmethod primary-button-pressed (editor figure x y)
  (declare (ignore figure x y))
  (:when (deselecting figure editor))
  (deselect-figures editor editor))


;;;;;;;;;;;;;;;;;;;;;;;;;
; PRIMARY-BUTTON-MOTION ;
;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric primary-button-motion (editor figure x y)
  (:predicates (moving (figure editor)
                 (and (not (brush-activep editor)) (drag-status editor)
                      (not (resize-activep editor))
                      (eql editor (figure-user figure))))
               (resizing (figure editor)
                 (and (not (brush-activep editor)) (drag-status editor)
                      (resize-activep editor)
                      (eql editor (figure-user figure))))
               (drawing (figure editor)
                 (and (brush-activep editor) (not (drag-status editor))))
               (drawing-selection (editor)
                 (and (not (brush-activep editor)) 
                      (not (drag-status editor))))))

(defmethod primary-button-motion (editor figure x y)
  (setf (drag-status-x (drag-status editor)) x)
  (setf (drag-status-y (drag-status editor)) y))

(defmethod primary-button-motion (editor figure x y)
  (:when (drawing figure editor))
  (draw-line editor x y))

(defmethod primary-button-motion (editor figure x y)
  (:when (moving figure editor))
  (let ((status (drag-status editor)))
    (move-figure editor figure 
                 (- x (drag-status-x status))
                 (- y (drag-status-y status)))
    (call-next-method)))

(defmethod primary-button-motion (editor figure x y)
  (:when (resizing figure editor))
  (let ((status (drag-status editor)))
;    (setf (figure-version figure) (1+ (figure-version figure)))
    (resize-figure editor figure 
                   (- x (drag-status-x status))
                   (- y (drag-status-y status)))
    (call-next-method)))

(defmethod primary-button-motion (editor figure x y)
  (:when (drawing-selection editor))
  (draw-selection-rectangle editor x y))

;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PRIMARY-BUTTON-RELEASED ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric primary-button-released (editor figure)
  (:predicates (resizing (figure editor)
                 (and (not (brush-activep editor)) (drag-status editor)
                      (resize-activep editor)))
               (drawing (figure editor)
                 (and (current-line editor) (brush-activep editor)))
               (drawing-selection (editor)
                 (selection-rectangle editor))))

(defmethod primary-button-released (editor figure)
  (setf (drag-status editor) nil)
  (gp:invalidate-rectangle editor))

(defmethod primary-button-released (editor figure)
  (:when (drawing figure editor))
  (end-drawing-line editor))

(defmethod primary-button-released (editor figure)
  (:when (resizing figure editor))
  (end-resizing-figure editor)
  (call-next-method))

(defmethod primary-button-released (editor figure)
  (:when (drawing-selection editor))
  (end-drawing-selection-rectangle editor))

