;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC CAPI EXTENSION
; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

(defclass button-pinboard-pane (capi:pinboard-layout)
  ((selected-button-position :initform 0 :initarg selected-button-position
                             :accessor selected-button-position)))

(defclass checkable-button (capi:image-pinboard-object)
  ((checked :initform nil :initarg :checkedp :accessor checkedp)
   (checked-image-location :initarg :checked-image-location
                           :accessor checked-image-location)
   (unchecked-image-location :initarg :unchecked-image-location
                             :accessor unchecked-image-location)
   (when-checked-lambda :initarg :when-checked :accessor when-checked-lambda)
   (when-unchecked-lambda :initarg :when-checked :accessor when-unchecked-lambda)))

(defmethod initialize-instance :after ((pane button-pinboard-pane) &key
                                       &allow-other-keys)
  (setf (capi:output-pane-input-model pane)
        `(((:button-1 :press) ,(lambda (current-pane x y)
                                 (let ((object (capi:pinboard-object-at-position current-pane x y)))
                                   (when object
                                     (change-check-status object))))))))

(defmethod initialize-instance :after ((button checkable-button) &key
                                       &allow-other-keys)
  (update-image button))

(defmethod change-check-status ((button checkable-button) &key (status nil status-supplied-p))
  (if status-supplied-p
       (setf (checkedp button) status)
    (unless (checkedp button)
      (setf (checkedp button) (not (checkedp button)))))
  (let* ((board (capi:pinboard-object-pinboard button))
         (interface (capi:top-level-interface board)))
    (if (checkedp button)
        (progn
          (uncheck-rest button)
          (funcall (fdefinition (when-checked-lambda button)) interface))
      (funcall (fdefinition (when-unchecked-lambda button)) interface))
    (update-image button)))

(defmethod uncheck-rest ((selected-button checkable-button))
  (let* ((pane (capi:pinboard-object-pinboard selected-button))
         (buttons (capi:layout-description pane)))
    (dolist (button buttons)
      (unless (eql button selected-button)
        (change-check-status button :status nil)))))

(defmethod update-image ((button checkable-button))
  (setf (capi:image-pinboard-object-image button)
        (if (checkedp button)
            (checked-image-location button)
          (unchecked-image-location button))))

; Admitted: The following method is to avoid some of the potential race conditions
; when using CAPI. The method below replaces the consecutive calls to pinboard-pane-size and
; pinboard-pane-position. Race conditions appear when these calls are made on an object that
; has a programmer-defined draw-pinboard-object method. 
(defmethod set-pinboard-object-geometry ((object capi:pinboard-object) &key x y width height)
  (let ((geometry (capi::geometry-cache object)))
    (setf (slot-value geometry 'capi-internals:height) height)
    (setf (slot-value geometry 'capi-internals:width) width)
    (setf (slot-value geometry 'capi::x) x)
    (setf (slot-value geometry 'capi::y) y)
    (gp:invalidate-rectangle (capi:pinboard-object-pinboard object))))

