;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
;
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC CLASS DEFINITIONS
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

; PREDICATED-METHOD method class
(defclass predicated-method (standard-method)
  ((predicates :initarg :predicates :initform nil :accessor method-predicates)))

; PREDICATED-FUNCTION generic function class
(defclass predicated-function (standard-generic-function)
  ((predicates :initarg :predicates :initform nil :accessor function-predicates))
  (:metaclass funcallable-standard-class))

; GROUP-FUNCTION generic function class
(defclass group-function (predicated-function)
  () (:metaclass funcallable-standard-class))

; PEER-FUNCTION generic function class
(defclass peer-function (predicated-function)
  ((group-function :initarg :group-function :initform nil :accessor peer-group-function))
  (:metaclass funcallable-standard-class))

; FUTURISED-FUNCTION generic function class
(defclass futurised-function (peer-function) ()
  (:metaclass funcallable-standard-class))

; GROUP-CLASS class
(define-distributable-class group-class ()
  ((peers :initform nil :initarg :members :accessor all-peers)
   (identity-function :initform 'same-class :initarg :identity-function
                      :accessor identity-function)))

(in-package :lambic)

(defvar *max-predicates* 10)
(defvar *group-level-call* nil)

; Root class for lambic predicates
(defclass lambic-predicate () ())

; TODO: to make it more dynamic (after PhD)
(defvar *lambic-predicate-classes*
  (list
   (defclass lambic-predicate-0 (lambic-predicate) ())
   (defclass lambic-predicate-1 (lambic-predicate-0) ())
   (defclass lambic-predicate-2 (lambic-predicate-1) ())
   (defclass lambic-predicate-3 (lambic-predicate-2) ())
   (defclass lambic-predicate-4 (lambic-predicate-3) ())
   (defclass lambic-predicate-5 (lambic-predicate-4) ())
   (defclass lambic-predicate-6 (lambic-predicate-5) ())
   (defclass lambic-predicate-7 (lambic-predicate-6) ())
   (defclass lambic-predicate-8 (lambic-predicate-7) ())
   (defclass lambic-predicate-9 (lambic-predicate-8) ())))
