;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC PACKAGE
;
; In Lambic, all the redefinitions of Common Lisp functions,
; methods, macros and special operators, are done in the 
; LAMBIC-USER package (lambic-redefinitions.lisp file), which
; is loaded only at the end (by the ASDF file). Before the
; execution of this file, such expressions use the original
; definitions. Their new definitions have different names so
; that they do not conflict with the original ones. For exapmle
; for the redefined operator PROGN, inside the Lambic package
; we use FUTURISED-PROGN.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defpackage :lambic
  (:use :c2cl)
  (:export ; pass-by-reference
           :defprotoclass :far-reference :reference-object-name
           ; Auxiliars
           :remove-if* :shadow-symbols
           ; Actors
           :in-actor-of :in-actors-of :spawn-actor :kill-actor 
           :define-distributable-class :actor-of
           ; The two symbols below are available to circumvent the
           ; restriction to invoke macros using funcall or apply.
           :send-to-actor-of :when-resolvedf
           ; Futures
           :future :when-resolved :resolve-with-result
           :resolve-with-discard-result :ruin-with-condition
           :resolvedp :future-result
           ; Communication (channels included)
           :when-self-disconnected :when-self-reconnected
           :when-disconnected :when-reconnected :start-communication
           :generic-channel :channel-reference :find-channel :provide-channel
           ; Discovery
           :publish-service :whenever-discovered :group-whenever-discovered
           ; Macros
           :defgroupclass :defgroupmethod :defgroupgeneric
           :define-peer-class :define-peer-generic :define-peer-method
           ; Sequentiality (library included)
           :futurise-functions :when-all-resolved
           :when-any-resolved :when-first-resolved :futurised-reduce
           :futurised-dolist :futurised-if :futurised-when
           :futurised-unless :futurised-handler-case
           :futurised-or :futurised-and :futurised-progn
           :futurised-remove-if :futurised-lambda
           :futurised-apply :futurised-funcall :futurised-find-if
           :parallel-progn :parallel-and :parallel-or
           :find-cache-function-future :find-cache-function-futures
           :*cache-provider-functions* :*cache-observer-functions*
           ; Context predicates
           :method-predicates :function-predicates
           ; Propagation
           :peer-group-function :group-peers :add-peer :find-peer :is-peer
           :identity-function :peer-discovered :peer-disconnected 
           :peer-reconnected :local-disconnection :local-reconnection
           :same-class :group-predicates :set-group-predicates :peer-position
           :first-in :each :default-reduce-results-to :peer-equal :this
           :*original-arguments* :*original-argument-names* :get-original-binding
           ; Global variables (used for debugging)
           :*device-address* :*device-port* :*distributable-objects*
           :*far-references* :*screen-output* :*current-actor*
           :*discovery-actor* :*communication-actor* :debug-print
           ; GUI-extension
           :button-pinboard-pane :checkable-button :change-check-status
           :uncheck-rest :update-image :set-pinboard-object-geometry
           ; Growl interaction
           :growl))

