;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC REDEFINITIONS
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#|

(in-package :lambic-user)

(defclass c () ((f :accessor c-f)))

(defmethod foo ((o c)) (print "foo"))

|#

(in-package :lambic-user)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; REDEFINITIONS OF MACROS AND SPECIAL OPERATORS ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(shadow-symbols '(progn defclass defgeneric defmethod dolist
                   if when unless or and lambda))

; PROGN
(defmacro progn (&rest args)
  `(lambic:futurised-progn ,@args))

; DEFCLASS
(defmacro defclass (&rest args)
  `(lambic:define-peer-class ,@args))

; DEFGENERIC
(defmacro defgeneric (&rest args)
  `(lambic:define-peer-generic ,@args))

; DEFMETHOD
(defmacro defmethod (&rest args)
  `(lambic:define-peer-method ,@args))

; IF
(defmacro if (&rest args)
  `(lambic:futurised-if ,@args))

; WHEN
(defmacro when (&rest args)
  `(lambic:futurised-when ,@args))

; UNLESS
(defmacro unless (&rest args)
  `(lambic:futurised-unless ,@args))
#|
; HANDLER-CASE
(defmacro handler-case (expression &rest cases)
  `(lambic:futurised-handler-case ,expression ,@cases))
|#

; OR
(defmacro or (&rest args)
  `(lambic:futurised-or ,@args))

; AND
(defmacro and (&rest args)
  `(lambic:futurised-and ,@args))

; LAMBDA
(defmacro lambda (&rest forms)
   `(lambic:futurised-lambda ,@forms))

; DOLIST
(defmacro dolist (&rest args)
  `(lambic:futurised-dolist ,@args))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; EXTRA MACROS TO GET LAMBIC PUBLICATIONS ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; BEGIN
; (like in Scheme)
(defmacro begin (&rest args)
  `(lambic:futurised-progn ,@args))

; PARALLEL-BEGIN
(defmacro parallel-begin (&rest args)
  `(lambic:parallel-progn ,@args))

; FOREACH
; (like in Scheme)
(defmacro foreach (&rest args)
  `(lambic:futurised-dolist ,@args))

; TRY-CATCH
; (like in Java)
(defmacro try-catch (&rest args)
  `(lambic:futurised-handler-case ,@args))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; REDEFINITIONS OF FUNCTIONS ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; A common reaction is to futurise all the available functions
; automatically. Unfortunately, in the current implementations 
; of Common Lisp, there is no easy way to do this. In addition, 
; the futurisation is not always necessary (or even desired). For
; example, the LIST function should not necessarily wait for the
; list's elements to be resolved before returning the list.
; Although this is not wrong, it may lead to unnecessary delays.
;
; Hence, thus far, we opt for manually choosing the functions to
; be futurised. This is something to be further evaluated.

; SETHASH
(define-peer-method sethash (var hash value)
  (setf (gethash var hash) value))

(shadow-symbols '(print format not + - map maphash gethash find find-if nth
                        reduce remove-if  apply funcall equal eql gp:draw-rectangle))

(futurise-functions '(print format (not :from :c2cl) + - map find
                            maphash gethash equal eql nth (draw-rectangle :from :gp)))

(define-peer-method reduce (function sequence)
   (lambic:futurised-reduce function sequence))

(define-peer-method remove-if (test sequence)
   (lambic:futurised-remove-if test sequence))


(define-peer-method find-if (test sequence)
   (lambic:futurised-find-if test sequence))

(define-peer-method apply (future-function args)
   (lambic:futurised-apply future-function args))

(c2cl:defmethod funcall (future-function &rest args)
   (apply future-function args))

; DISPLAY
(c2cl:defmethod display (&rest args)
  (apply (lambda (args)
           (dolist (arg args)
             (princ arg *screen-output*)))
         args))



; Make SETHASH the "setf" of GETHASH.
(defsetf gethash sethash)
